/*
 * cobrowse
 * copyright (c) 2005 mark l. chang
 * http://faculty.olin.edu/~mchang
 * http://cobrowse.mozdev.org
 * 
 * This extension was designed to work with the 1.0+ series of
 * Firefox. Works with 2.0+ series after minimal testing.
 *
 * Essentially, this extension is a collaborative web browser.
 * "Collaborative" is a bit of a stretch, so we should actually think
 * of it in a leader/follower sense. With cobrowse installed, one
 * user can start a "server" that will accept incoming connections 
 * (currently on port 5000) from folks that have chosen to connect
 * via "client mode". 
 * 
 * The server browser then sends every web page navigation change
 * to every client. It even works when changing tabs!
 *
 * This code borrows heavily from these resources:
 * Google Maps Live!
 *  http://ted.mielczarek.org/code/mozilla/gmaplive/
 *  I borrowed a lot of the client/server model code from here
 * Firefox toolbar tutorial
 *  http://www.borngeek.com/firefox/tutorial/
 *  Just awesome. Thanks!
 * Roachfiend's "how to create firefox extensions" tutorial
 *  http://roachfiend.com/archives/2004/12/08/how-to-create-firefox-extensions/
 *  I got my first template/bundle here! Thanks!
 * XUL Planet
 *  http://www.xulplanet.com/
 *  The reference you need.
 *
 * CHANGELOG
 *  25-oct-2006:
 *   0.17 - made compatible with FF 2.0. Just changed version control info in install.rdf
 *  04-dec-2005:
 *   0.16 - made compatible with FF 1.5. Just changed version control info in install.rdf
 *  14-nov-2005:
 *   0.15 - removed the update url pointing to olin.edu. mozilla does not
 *          approve using external (non addons.mozilla.org) urls.
 *          no functionality changes.
 *  13-nov-2005:
 *   0.14 - added cut and paste of your hostname, also connected client count
 *  12-nov-2005:
 *   --   - mcolyer added ip lookup and a button to copy to clipboard
 *   0.13 - oh boy, i fuggered up the packaging. that's what messes it all up!
 *          you may have to manually uninstall your current version... sorry!
 *   0.12 - oh man, i got internal versions wrong
 *  11-nov-2005:
 *   0.11 - added updating capability
 *  11-nov-2005:
 *   0.1 - first release. v
 *
 * TODO
 *  actual real error handling :)
 *  make preferences work
 *  user-selectable ports
 *  new graphics
 */

var client;
var server;
var obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);

// register our fancy progress event listener
addEventListener("load", onInit, false);
addEventListener("unload", onUninit, false);
    
function onInit() {
  gBrowser.addProgressListener(myBrowserProgressListener);
}
    
function onUninit() {
  gBrowser.removeProgressListener(myBrowserProgressListener);
}

// the progress listener
var myBrowserProgressListener = {
    stateIsRequest:false,
    
    QueryInterface : function(aIID) {
        if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports))
                return this;
        throw Components.results.NS_NOINTERFACE;
    },
    
    onStateChange : function(a,b,c,d) {},
    onProgressChange : function(a,b,c,d,e,f){},
    onStatusChange : function(a,b,c,d){},
    onSecurityChange : function(a,b,c){},
    onLinkIconAvailable : function(a){},
    
    onLocationChange : function(aProgress, aRequest, aLocation) {
        onBrowserLocationChange(aLocation);
    }
};

// on a browser location change, update the observers
// this is *it*
function onBrowserLocationChange(aLocation) {
   //alert("New URL is " + aLocation.spec.toString());
   var newUrl = aLocation.spec.toString() + '\n';
   obs.notifyObservers(null,"cobrowse-newurl",newUrl);
}

// function to instantiate a server object
// does NOT start the server
// the server gets started and stopped, but never deleted
function cobrowseServer() {
    cobrowse_log("creating a cobrowseServer");
    
    this.serverSocket = null;
    this.serverRunning = false;
    this.numClients = 0;
    
    // observe for shutdown
    obs.addObserver(this,"quit-application",false);
}

// defintes a server class
cobrowseServer.prototype = {
    // start the server listening on port 5000
    // fix to make it flexible port
    startServer : function() {
        if(this.serverSocket)
            return;

        try {
            cobrowse_log("starting cobrowseServer");
            this.serverSocket = Components.classes["@mozilla.org/network/server-socket;1"].createInstance(Components.interfaces.nsIServerSocket);
            this.serverSocket.init(5000,false,-1);
            this.serverSocket.asyncListen(this);
            this.serverRunning = true;
        } catch(ex) {
            dump(ex + "\n");
        }
    },

    // stops teh server, but does not delete it
    // mostly because we can't delete ourselves :)
    stopServer : function() {
        if(this.serverSocket) {
            cobrowse_log("trying to stop server socket");
            try {
                this.serverSocket.close();
            }
            catch(ex) {}
            this.serverSocket = null;
            this.serverRunning = false;
            this.numClients = 0;
            obs.notifyObservers(null,"cobrowse-stopserver",null);
        }
    },

    // nsIServerSocketListener functions
    
    // create a new stream listener when we get a new connection
    // each stream listener is actually in output mode to send
    // information to the client
    onSocketAccepted : function onSocketAccepted(serverSocket, transport) {
        cobrowse_log("got a connection");

        var ostream = transport.openOutputStream(0,0,0);
        var istream = transport.openInputStream(0,0,0);
        if( ostream && istream ) {
            // create bidirectional streams in the cobrowseWriter class
            var cbwriter = new cobrowseWriter(ostream);
            cbwriter.inputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
            cbwriter.inputStream.init(istream);
            cbwriter.parentServer = this;
            var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"].createInstance(Components.interfaces.nsIInputStreamPump);
            pump.init(istream, -1, -1, 0, 0, false);
            pump.asyncRead(cbwriter, null);
        }
        
        // update the UI to reflect number of clients
        this.addClient();
    },
    
    // stop the serversocket. does not stop the individual streams
    onStopListening : function onStopListening(serverSocket, status) {
        this.serverSocket = null;
        this.serverRunning = false;
    },
    // end nsIServerSocketListener
    
    // adds a client to the UI
    addClient: function() {
        this.numClients++;
        var serveButton = document.getElementById("cobrowse-serve-button");
        serveButton.label = "Stop (" + this.numClients + ") ";    
    },
    
    // removes a client from the UI
    removeClient: function(myCobrowseClient) {
        myCobrowseClient = null; // does this actually get garbage collected?
        this.numClients--;
        var serveButton = document.getElementById("cobrowse-serve-button");
        serveButton.label = "Stop (" + this.numClients + ") ";
    },
    
    // nsIObserver
    // let's us stop the server before quitting mozilla
    observe: function(subject, topic, data) {
        if(topic == "quit-application") {
            this.stopServer();
            obs.removeObserver(this,"quit-application");
        }
    }
}

// creates a writer stream to clients
function cobrowseWriter(stream) {
    cobrowse_log("starting a writer");
    this.outputStream = stream;
    // register with the observer service
    obs.addObserver(this,"cobrowse-newurl",false);
    obs.addObserver(this,"cobrowse-stopserver",false);
    obs.addObserver(this,"quit-application",false);
}

// class definition of the writer object
cobrowseWriter.prototype = {
    // close up
    disconnect: function() {
        cobrowse_log("closing writer stream");
        try {
            this.outputStream.close();
            this.inputStream.close();
            this.outputStream = null;
            this.inputStream = null;
        } catch(e) { dump(e); }
        // unregister from the service
        obs.removeObserver(this, "cobrowse-newurl");
        obs.removeObserver(this, "cobrowse-stopserver");
        obs.removeObserver(this, "quit-application");
        // update the parent (for the UI)
        this.parentServer.removeClient(this);
    },
    
    // write output to client
    send: function(msg) {
        cobrowse_log("writing " + msg);
        try {
            this.outputStream.write(msg, msg.length);
            this.outputStream.flush();
        } catch(e) { dump(e); }
    },
    
    // nsIStreamListener methods
    onStartRequest: function(request, context) {},
    onStopRequest: function(request, context, status) {
        cobrowse_log("cb_writer: onStopRequest");
        if( this.inputStream ) {
            try {
                cobrowse_log("Shutting down a cbwriter input stream");
                this.inputStream.close();
                this.inputStream = null;
            } catch (e) { dump(e); }
        }
        this.disconnect();
    },
    
    onDataAvailable: function(request, context, inStream, offset, count) {},
    
    observe: function(subject, topic, data) {
        if( topic == "cobrowse-newurl" ) {
            cobrowse_log("got newurl");
            this.send(data);
        }
        if( topic == "cobrowse-stopserver" ) {
            cobrowse_log("got stopserver");
            this.disconnect();
        }
        if( topic == "quit-application" ) {
            cobrowse_log("got quit");
            this.disconnect();
        }
    }
}

// this cobrowseClient object takes a server string argument
// and attempts to connect to it using a socket transport
// the transportService, upon timeout/error, will call the onStopRequest
// at the pump level to disconnect the streams
function cobrowseClient(server) {
    try {
        cobrowse_log("trying to connect to " + server);
        this.transportService = Components.classes["@mozilla.org/network/socket-transport-service;1"].getService(Components.interfaces.nsISocketTransportService);
        this.transport = this.transportService.createTransport(null, 0, server, 5000, null);
        this.outputStream = this.transport.openOutputStream(0, 0, 0);
        this.stream = this.transport.openInputStream(0, 0, 0);
        this.inputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
        this.inputStream.init(this.stream);

        cobrowse_log("starting pump");
        var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"].createInstance(Components.interfaces.nsIInputStreamPump);
        pump.init(this.stream, -1, -1, 0, 0, false);
        pump.asyncRead(this, null);

        cobrowse_log("done");
    } catch (e) {
        alert("Error starting client! Bailing.");
        return null;
    }
}

// the definition of the cobrowseClient class
cobrowseClient.prototype = {
    // nsIStreamListener methods
    onStartRequest: function(request, context) {},

    onStopRequest: function(request, context, status) {
        cobrowse_log("onStopRequest");
        if( status == "userquit" ) {
            // user called it quits
            cobrowse_log("user closed connection");
            var msg = "GOODBYE";
            this.outputStream.write(msg, msg.length);
        }
        else {
            // server closed us
            cobrowse_log("server closed connection");
        }
        
        if( this.inputStream ) {
            try {
                cobrowse_log("Shutting down input stream");
                this.inputStream.close();
                this.inputStream = null;
            } catch (e) { dump(e); }
        }
        if( this.outputStream ) {
            try {
                cobrowse_log("Shutting down output stream");
                this.outputStream.close();
                this.outputStream = null;
            } catch (e) { dump(e); } 
        }
        cobrowse_stopClient();
    },

    // grabs data and changes URLs as we get them
    onDataAvailable: function(request, context, inStream, offset, count) {
        inputText = this.inputStream.read(count);
        cobrowse_log( inputText );
        this.loadUrl( inputText );
    },

    // called when the user clicks on the disconnect button
    disconnect: function() {
        cobrowse_log("disconnect");
        this.onStopRequest(null, null, "userquit" );
    },

    loadUrl: function( url ) {
        // Set the document's location to the incoming URL
        window._content.document.location = url;

        // Make sure we get the focus
        window.content.focus();
    }
}

// called by the user interface button
// copies the local hostname into the
// system clipboard to make it easier to join a session
function cobrowse_copyhostname() {
    cobrowse_log("Copying hostname to clipboard");
    
    var dns = Components.classes['@mozilla.org/network/dns-service;1'].getService(Components.interfaces.nsIDNSService);
    var ip = dns.myHostName;
    
    var oClipBoard = Components.classes["@mozilla.org/widget/clipboardhelper;1"].getService(Components.interfaces.nsIClipboardHelper);
    oClipBoard.copyString(ip);
}

// called by the user interface button
// starts and stops the server interface
// creates a new object to handle each new connection (stream)
// each object is registered with an observable, so that
// we can send data to all of the stream endpoints
// when a new URL is loaded on the server side
function cobrowse_toggleServer() {
    if( !server ) {
        server = new cobrowseServer();
        server.startServer();
        // disable client
        var clientButton = document.getElementById("cobrowse-client-button");
        var serveButton = document.getElementById("cobrowse-serve-button");
        clientButton.disabled = true;
        serveButton.label = "Stop (0) ";
    } else if ( server ) {
        server.stopServer();
        server = null;
        // enable client and server
        var clientButton = document.getElementById("cobrowse-client-button");
        var serveButton = document.getElementById("cobrowse-serve-button");
        clientButton.disabled = false;
        serveButton.label = "Serve ";
    }
}


// this does not actually STOP the client socket interface
// called from the onStopRequest pump callback, if the server side
// closes, we disconnect and then call this to update the
// client object (delete it) and update the user interface buttons
function cobrowse_stopClient() {
    client = null;
    var clientButton = document.getElementById("cobrowse-client-button");
    var serveButton = document.getElementById("cobrowse-serve-button");
    clientButton.label = "Connect ";
    serveButton.disabled = false;
}

// switches the client on and off
// called by the connect button in the toolbar
function cobrowse_toggleClient() {
    if( !client ) { // start the client
        // get a handle to the server name
        var serverNameBox = document.getElementById("cobrowse-server-name");

        // get the value in the server name box, trimming as necessary
        var serverName = cobrowse_trimString(serverNameBox.value);

        // if there was no server, we can't start
        if( serverName.length != 0 ) {
            cobrowse_addServer(serverName);

            // try connecting
            client = new cobrowseClient(serverName);

            // success? change connect to stop, server to disabled
            if( client ) {
                var clientButton = document.getElementById("cobrowse-client-button");
                var serveButton = document.getElementById("cobrowse-serve-button");
                clientButton.label = "Stop ";
                serveButton.disabled = true;
            }
            else {
                // oops, something didn't work
                alert("Couldn't connect to " + serverName);
            }
        }
        else {
            alert("Please set server name.");
        }
    }
    else if (client) { // stop the client
        client.disconnect();
        client = null;
        var clientButton = document.getElementById("cobrowse-client-button");
        var serveButton = document.getElementById("cobrowse-serve-button");
        clientButton.label = "Connect";
        serveButton.disabled = false;
    }
}

// ---------------- UI / utility functions -------------------

const cobrowse_consoleService =
      Components.
          classes['@mozilla.org/consoleservice;1'].
              getService(Components.interfaces.nsIConsoleService);

// adds a server that you have accessed to the drop down
// list of servers in the toolbar
function cobrowse_addServer(serverName) {
    // get the server menu
    var menu = document.getElementById("cobrowse-server-menu");
    var serverItem = document.createElement("menuitem");
    serverItem.setAttribute("label", serverName);
    menu.appendChild(serverItem);
}

// populates the server list drop down menu with some
// defaults. used mainly for ease of debugging now,
// but should really come out of the preferences
function cobrowse_populateServerList() {
//     // get the menu element we want to populate
//     var menu = document.getElementById("cobrowse-server-menu");

//     // clean out the menu
//     for (var i=menu.childNodes.length - 1; i>=0; i--) {
//         menu.removeChild(menu.childNodes.item(i));
//     }

//     // put in a default server list, just for kicks
//     var serverItem = document.createElement("menuitem");
//     serverItem.setAttribute("label", "fpga.olin.edu");

//     // add it to the item
//     menu.appendChild(serverItem);
}

// trims a string
function cobrowse_trimString(string)
{
    // Return empty if nothing was passed in
    if (!string) return "";

    // Efficiently replace any leading or trailing whitespace
    var value = string.replace(/^\s+/, '');
    value = string.replace(/\s+$/, '');

    // Replace any multiple whitespace characters with a single space
    value = value.replace(/\s+/g, ' ');

    // Return the altered string
    return value;
}

// popups an alert with static text
function cobrowse_showAlert() {
    alert("CoBrowsing is FUN");
}

// sends text to the javascript console log
function cobrowse_log(msg) {
    cobrowse_consoleService.logStringMessage(msg);
}

// called by the debug user interface element
// does nothing now
function cobrowse_debug() {
//     // Set the document's location to the incoming URL
//     window._content.document.location = "http://acl.olin.edu";

//     // Make sure we get the focus
//     window.content.focus();
    obs.notifyObservers(null,"cobrowse-newurl","http://www.google.com");
}
